"""
Exchange stream cache for market and order data.

This cache implementation tries to be efficient and minimalistic by using the following principles:
- Use builtin data structures whenever possible
- Minimize creation and allocation of new objects
- Avoid any kind of sorting

Though, there is probably quite some room for further optimizations.
"""

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Generic, NamedTuple, TypeAlias, TypeVar

from betfair_parser.spec.streaming import (
    LPV,
    MCM,
    OCM,
    PV,
    ChangeType,
    MarketDefinition,
    MatchedOrder,
    Order,
    OrderRunnerChange,
    RunnerChange,
    RunnerDefinition,
    SegmentType,
)


class SelectionKey(NamedTuple):
    selection_id: int
    handicap: float


def get_selection_key(selection: RunnerChange | OrderRunnerChange | RunnerDefinition) -> SelectionKey:
    """Asian handicap markets use a combined key to represent all selections in a market."""
    return SelectionKey(selection.id, selection.hc or 0.0)


V = TypeVar("V")
SelectionKeyTypes: TypeAlias = int | tuple[int, float] | SelectionKey


class SelectionDict(defaultdict[SelectionKey, V], Generic[V]):
    """
    A defaultdict whose real key type is SelectionKey,
    but which also lets you do d[5] or d[(5,1.0)].
    """

    default_factory: type[V]

    def __init__(self) -> None:
        super().__init__(self.default_factory)

    @staticmethod
    def _normalize(key: SelectionKeyTypes) -> SelectionKey:
        if isinstance(key, SelectionKey):
            return key
        if isinstance(key, tuple):
            return SelectionKey(*key)
        if isinstance(key, int):
            return SelectionKey(key, 0.0)
        raise TypeError(f"Invalid key type: {type(key).__name__}")

    def __getitem__(self, key: SelectionKeyTypes) -> V:
        return super().__getitem__(self._normalize(key))

    def __contains__(self, key: SelectionKeyTypes) -> bool:  # type: ignore[override]
        return super().__contains__(self._normalize(key))

    def __setitem__(self, key: SelectionKeyTypes, value: V) -> None:
        return super().__setitem__(self._normalize(key), value)

    def __delitem__(self, key: SelectionKeyTypes) -> None:
        return super().__delitem__(self._normalize(key))

    def get(self, key: SelectionKeyTypes, default: V | None = None) -> V | None:  # type: ignore[override]
        return super().get(self._normalize(key), default)


PriceVolumeMap: TypeAlias = dict[float, float]
LadderPriceVolumeMap: TypeAlias = dict[int, LPV]


def ladder_update_lpv(ladder: LadderPriceVolumeMap, lpvs: list[LPV]) -> None:
    for lpv in lpvs:
        if not lpv.volume:
            ladder.pop(lpv.level, None)
        else:
            ladder[lpv.level] = lpv


def ladder_update_pv(ladder: PriceVolumeMap, pvs: list[PV]) -> None:
    for pv in pvs:
        if not pv.volume:
            ladder.pop(pv.price, None)
        else:
            ladder[pv.price] = pv.volume


def ladder_update_mo(ladder: PriceVolumeMap, mos: list[MatchedOrder]) -> None:
    for mo in mos:
        if not mo.size:
            ladder.pop(mo.price, None)
        else:
            ladder[mo.price] = mo.size


class BestPriceVolume(NamedTuple):
    back_price: float
    back_volume: float
    lay_price: float
    lay_volume: float


_NO_PRICE = 0.0  # should this be None?
_NO_PRICE_LPV = LPV(0, _NO_PRICE, 0.0)


@dataclass(slots=True)
class RunnerOrderBook:
    """Order book for a runner with all relevant price and volume levels."""

    available_to_back: PriceVolumeMap = field(default_factory=dict)
    available_to_lay: PriceVolumeMap = field(default_factory=dict)

    best_available_to_back: LadderPriceVolumeMap = field(default_factory=dict)
    best_available_to_lay: LadderPriceVolumeMap = field(default_factory=dict)

    best_display_available_to_back: LadderPriceVolumeMap = field(default_factory=dict)
    best_display_available_to_lay: LadderPriceVolumeMap = field(default_factory=dict)

    starting_price_back: PriceVolumeMap = field(default_factory=dict)
    starting_price_lay: PriceVolumeMap = field(default_factory=dict)

    starting_price_near: float | None = None
    starting_price_far: float | None = None

    traded: PriceVolumeMap = field(default_factory=dict)
    last_traded_price: float | None = None
    total_volume: float | None = None

    def update(self, rc: RunnerChange) -> None:
        if rc.atb:
            ladder_update_pv(self.available_to_back, rc.atb)
        if rc.atl:
            ladder_update_pv(self.available_to_lay, rc.atl)
        if rc.batb:
            ladder_update_lpv(self.best_available_to_back, rc.batb)
        if rc.batl:
            ladder_update_lpv(self.best_available_to_lay, rc.batl)
        if rc.bdatb:
            ladder_update_lpv(self.best_display_available_to_back, rc.bdatb)
        if rc.bdatl:
            ladder_update_lpv(self.best_display_available_to_lay, rc.bdatl)
        if rc.spb:
            ladder_update_pv(self.starting_price_back, rc.spb)
        if rc.spl:
            ladder_update_pv(self.starting_price_lay, rc.spl)
        if rc.spn:
            self.starting_price_near = rc.spn
        if rc.spf:
            self.starting_price_far = rc.spf
        if rc.trd:
            ladder_update_pv(self.traded, rc.trd)
        if rc.ltp:
            self.last_traded_price = rc.ltp
        if rc.tv:
            self.total_volume = rc.tv

    def best_prices(self) -> BestPriceVolume:
        """Extract the best available prices."""

        if self.available_to_back or self.available_to_lay:
            back_price = max(self.available_to_back) if self.available_to_back else _NO_PRICE
            back_volume = self.available_to_back[back_price] if back_price else 0.0
            lay_price = min(self.available_to_lay) if self.available_to_lay else _NO_PRICE
            lay_volume = self.available_to_lay[lay_price] if lay_price else 0.0
            return BestPriceVolume(back_price, back_volume, lay_price, lay_volume)

        if self.best_available_to_back or self.best_available_to_lay:
            blpv = self.best_available_to_back.get(0, _NO_PRICE_LPV)
            llpv = self.best_available_to_lay.get(0, _NO_PRICE_LPV)
            return BestPriceVolume(blpv.price, blpv.volume, llpv.price, llpv.volume)

        if self.best_display_available_to_back or self.best_display_available_to_lay:
            blpv = self.best_display_available_to_back.get(0, _NO_PRICE_LPV)
            llpv = self.best_display_available_to_lay.get(0, _NO_PRICE_LPV)
            return BestPriceVolume(blpv.price, blpv.volume, llpv.price, llpv.volume)

        return BestPriceVolume(_NO_PRICE, 0.0, _NO_PRICE, 0.0)

    def __repr__(self) -> str:
        data: dict[str, str] = {}
        for key in self.__dataclass_fields__:
            val = getattr(self, key)
            if not val:
                continue

            if isinstance(val, dict):
                if key in ("available_to_back", "available_to_lay"):
                    formatted = ", ".join(f"{price:.4g}:{volume:.2f}" for price, volume in sorted(val.items()))
                else:
                    formatted = ", ".join(f"{lpv.price:.4g}:{lpv.volume:.2f}" for _, lpv in sorted(val.items()))
                val_str = f"[{formatted}]"
            elif isinstance(val, float):
                val_str = f"{val:.2f}"
            else:
                val_str = str(val)

            data[key] = val_str

        fields_str = ", ".join(f"{k}={v}" for k, v in sorted(data.items()))
        space = " " if fields_str else ""
        return f"<{type(self).__name__}{space}{fields_str}>"


class ChangeCache:
    clk: str | None = None
    initial_clk: str | None = None
    publish_time: int | None = None
    stream_unreliable: bool | None = False
    conflate_ms: int | None = None

    def update_meta(self, msg: MCM | OCM) -> None:
        if msg.initial_clk:
            self.initial_clk = msg.initial_clk
        if msg.clk:
            self.clk = msg.clk
        if msg.conflate_ms:
            self.conflate_ms = msg.conflate_ms
        self.publish_time = msg.pt
        self.stream_unreliable = msg.stream_unreliable
        if msg.ct == ChangeType.SUB_IMAGE and not msg.segment_type or msg.segment_type == SegmentType.SEG_START:
            self.clear()

    def clear(self) -> None:
        raise NotImplementedError()


class MarketOrderBook(SelectionDict[RunnerOrderBook]):
    """Order book for a single market, collecting a bunch of RunnerOrderBooks."""

    default_factory: type = RunnerOrderBook


class MarketSubscriptionCache(ChangeCache):
    """
    Orderbook for all markets of a market change stream.

    Market subscriptions are always in the underlying exchange currency - GBP. The default roll-up for GBP
    is £1 for batb / batl and bdatb / bdatl, This means that stakes of less than £1 (or currency
    equivalent) are rolled up to the next available price on the odds ladder. For atb / atl there is
    no roll-up. Available volume is displayed at all prices including those with less than £2 available.
    """

    def __init__(self):
        self.order_book: defaultdict[str, MarketOrderBook] = defaultdict(MarketOrderBook)
        self.definitions: dict[str, MarketDefinition] = {}

    def clear(self) -> None:
        self.order_book.clear()
        self.definitions.clear()

    def update(self, mcm: MCM) -> None:
        self.update_meta(mcm)
        if mcm.is_heartbeat:
            return
        if not mcm.mc:
            return

        for mc in mcm.mc:
            if mc.img:
                self.order_book.pop(mc.id, None)
            if mc.market_definition:
                self.definitions[mc.id] = mc.market_definition
            if not mc.rc:
                continue
            for rc in mc.rc:
                self.order_book[mc.id][get_selection_key(rc)].update(rc)


class RunnerOrders:
    """
    New subscriptions: Will receive an initial image with only E - Executable orders (unmatched).
    Live subscriptions: Will receive a transient of the order to EC - Execution Complete as the
    order transits into that state (allowing you to remove the order from your cache).

    Please note: EXECUTION_COMPLETE (fully matched) orders are only returned when transitioning from
    EXECUTABLE to EXECUTION_COMPLETE. The full details of EXECUTION_COMPLETE orders can only be viewed
    using listCurrentOrders/listMarketBook using orderProjections.
    """

    __slots__ = (
        "matched_backs",
        "matched_lays",
        "unmatched_orders",
        "executed_orders",
        "handicap",
    )

    def __init__(self) -> None:
        self.matched_backs: dict[float, float] = {}
        self.matched_lays: dict[float, float] = {}
        self.unmatched_orders: dict[int, Order] = {}
        self.executed_orders: dict[int, Order] = {}  # Does only contain recently completed orders

    def update(self, orc: OrderRunnerChange) -> None:
        if orc.uo:
            for uo in orc.uo:
                if uo.execution_complete:
                    self.unmatched_orders.pop(uo.id, None)
                    self.executed_orders[uo.id] = uo
                else:
                    self.unmatched_orders[uo.id] = uo
        if orc.mb:
            ladder_update_mo(self.matched_backs, orc.mb)
        if orc.ml:
            ladder_update_mo(self.matched_lays, orc.ml)


class MarketOrders(SelectionDict[RunnerOrders]):
    """All orders for a single market, collecting a bunch of RunnerOrders."""

    default_factory: type = RunnerOrders


class OrderSubscriptionCache(ChangeCache):
    """
    Order subscriptions are provided in the currency of the account that the orders are placed in.
    """

    def __init__(self):
        self.orders: defaultdict[str, MarketOrders] = defaultdict(MarketOrders)

    def clear(self) -> None:
        self.orders.clear()

    def update(self, ocm: OCM) -> None:
        self.update_meta(ocm)
        if ocm.is_heartbeat:
            return
        if not ocm.oc:
            return

        for oc in ocm.oc:
            if oc.full_image:
                self.orders.pop(oc.id, None)
            if oc.closed:
                self.orders.pop(oc.id, None)
            if not oc.orc:
                # on closed markets
                continue
            for orc in oc.orc:
                orc_key = get_selection_key(orc)
                if orc.full_image:
                    self.orders[oc.id].pop(orc_key, None)
                self.orders[oc.id][orc_key].update(orc)



from betfair_parser.strenums import DocumentedEnum, StrEnum, auto, doc


class MarketProjection(DocumentedEnum):
    COMPETITION = doc("If not selected then the competition will not be returned with marketCatalogue")
    EVENT = doc("If not selected then the event will not be returned with marketCatalogue")
    EVENT_TYPE = doc("If not selected then the eventType will not be returned with marketCatalogue")
    MARKET_START_TIME = doc("If not selected then the start time will not be returned with marketCatalogue")
    MARKET_DESCRIPTION = doc("If not selected then the description will not be returned with marketCatalogue")
    RUNNER_DESCRIPTION = doc("If not selected then the runners will not be returned with marketCatalogue")
    RUNNER_METADATA = doc(
        "If not selected then the runner metadata will not be returned with marketCatalogue. If selected "
        "then RUNNER_DESCRIPTION will also be returned regardless of whether it is included as a market "
        "projection."
    )


class PriceData(DocumentedEnum):
    SP_AVAILABLE = doc("Amount available for the BSP auction.")
    SP_TRADED = doc("Amount traded in the BSP auction.")
    EX_BEST_OFFERS = doc("Only the best prices available for each runner, to requested price depth.")
    EX_ALL_OFFERS = doc("EX_ALL_OFFERS trumps EX_BEST_OFFERS if both settings are present")
    EX_TRADED = doc("Amount traded on the exchange.")


class MatchProjection(DocumentedEnum):
    NO_ROLLUP = doc("No rollup, return raw fragments")
    ROLLED_UP_BY_PRICE = doc("Rollup matched amounts by distinct matched prices per side.")
    ROLLED_UP_BY_AVG_PRICE = doc("Rollup matched amounts by average matched price per side")


class OrderProjection(DocumentedEnum):
    ALL = doc("EXECUTABLE and EXECUTION_COMPLETE orders")
    EXECUTABLE = doc(
        "An order that has a remaining unmatched portion. This is either a fully unmatched "
        "or partially matched bet (order)"
    )
    EXECUTION_COMPLETE = doc(
        "An order that does not have any remaining unmatched portion. This is a fully matched bet (order)."
    )


class MarketStatus(DocumentedEnum):
    INACTIVE = doc("The market has been created but isn't yet available.")
    OPEN = doc("The market is open for betting.")
    SUSPENDED = doc("The market is suspended and not available for betting.")
    CLOSED = doc("The market has been settled and is no longer available for betting.")


class RunnerStatus(DocumentedEnum):
    ACTIVE = auto()
    WINNER = auto()
    LOSER = auto()
    PLACED = doc("The runner was placed, applies to EACH_WAY marketTypes only.")
    REMOVED_VACANT = doc(
        "REMOVED_VACANT applies to Greyhounds. Greyhound markets always return a fixed number of "
        "runners (traps). If a dog has been removed, the trap is shown as vacant."
    )
    REMOVED = auto()
    HIDDEN = doc(
        "The selection is hidden from the market. This occurs in Horse Racing markets were runners "
        "is hidden when it is doesn't hold an official entry following an entry stage. This could be "
        "because the horse was never entered or because they have been scratched from a race at a "
        "declaration stage. All matched customer bet prices are set to 1.0 even if there are later "
        "supplementary stages. Should it appear likely that a specific runner may actually be supplemented "
        "into the race this runner will be reinstated with all matched customer bets set back to "
        "the original prices."
    )


class TimeGranularity(StrEnum):
    DAYS = auto()
    HOURS = auto()
    MINUTES = auto()


class Side(DocumentedEnum):
    BACK = doc(
        "To back a team, horse or outcome is to bet on the selection to win. For LINE markets a Back bet "
        "refers to a SELL line. A SELL line will win if the outcome is LESS THAN the taken line (price)."
    )
    LAY = doc(
        "To lay a team, horse, or outcome is to bet on the selection to lose. For LINE markets a Lay bet "
        "refers to a BUY line. A BUY line will win if the outcome is MORE THAN the taken line (price)."
    )


class OrderBy(DocumentedEnum):
    BY_BET = doc("@Deprecated Use BY_PLACE_TIME instead. Order by placed time, then bet id.")
    BY_MARKET = doc("Order by market id, then placed time, then bet id.")
    BY_MATCH_TIME = doc(
        "Order by time of last matched fragment (if any), then placed time, then bet id. Filters "
        "out orders which have no matched date. The dateRange filter (if specified) is applied to "
        "the matched date."
    )
    BY_PLACE_TIME = doc(
        "Order by placed time, then bet id. This is an alias of to be deprecated BY_BET. The "
        "dateRange filter (if specified) is applied to the placed date."
    )
    BY_SETTLED_TIME = doc(
        "Order by time of last settled fragment (if any due to partial market settlement), then by "
        "last match time, then placed time, then bet id. Filters out orders which have not been settled. "
        "The dateRange filter (if specified) is applied to the settled date."
    )
    BY_VOID_TIME = doc(
        "Order by time of last voided fragment (if any), then by last match time, then placed time, then "
        "bet id. Filters out orders which have not been voided. The dateRange filter (if specified) is "
        "applied to the voided date."
    )


class SortDir(DocumentedEnum):
    EARLIEST_TO_LATEST = doc("Order from earliest value to latest e.g. lowest betId is first in the results.")
    LATEST_TO_EARLIEST = doc("Order from the latest value to the earliest e.g. highest betId is first in the results.")


class MarketSort(DocumentedEnum):
    MINIMUM_TRADED = doc("Minimum traded volume")
    MAXIMUM_TRADED = doc("Maximum traded volume")
    MINIMUM_AVAILABLE = doc("Minimum available to match")
    MAXIMUM_AVAILABLE = doc("Maximum available to match")
    FIRST_TO_START = doc("The closest markets based on their expected start time")
    LAST_TO_START = doc("The most distant markets based on their expected start time")


class MarketBettingType(DocumentedEnum):
    ODDS = doc("Odds Market - Any market that doesn't fit any any of the below categories.")
    LINE = doc(
        "Line Market - LINE markets operate at even-money odds of 2.0. However, price for these markets "
        "refers to the line positions available as defined by the markets min-max range and interval steps. "
        "Customers either Buy a line (LAY bet, winning if outcome is greater than the taken line (price)) or "
        "Sell a line (BACK bet, winning if outcome is less than the taken line (price)). If settled outcome "
        "equals the taken line, stake is returned."
    )
    RANGE = doc("Range Market - Now Deprecated")
    ASIAN_HANDICAP_DOUBLE_LINE = doc(
        "Asian Handicap Market - A traditional Asian handicap market. Can be identified by marketType ASIAN_HANDICAP"
    )
    ASIAN_HANDICAP_SINGLE_LINE = doc(
        "Asian Single Line Market - A market in which there can be 0 or multiple winners. e.g. marketType TOTAL_GOALS"
    )
    FIXED_ODDS = doc(
        "Sportsbook Odds Market. This type is deprecated and will be removed in future releases, when "
        "Sportsbook markets will be represented as ODDS market but with a different product type."
    )


class ExecutionReportStatus(DocumentedEnum):
    SUCCESS = doc("Order processed successfully")
    FAILURE = doc("Order failed.")
    PROCESSED_WITH_ERRORS = doc(
        "The order itself has been accepted, but at least one (possibly all) actions have generated errors. "
        "This error only occurs for replaceOrders, cancelOrders and updateOrders operations. In normal "
        "circumstances the placeOrders operation will not return PROCESSED_WITH_ERRORS status as it is an "
        "atomic operation.  PLEASE NOTE: if the 'Best Execution' features is switched off, placeOrders can "
        "return ‘PROCESSED_WITH_ERRORS’ meaning that some bets can be rejected and other placed when "
        "submitted in the same PlaceInstruction"
    )
    TIMEOUT = doc(
        "The order timed out & the status of the bet is unknown. If a TIMEOUT error occurs on a "
        "placeOrders/replaceOrders request, you should check listCurrentOrders to verify the status of your "
        "bets before placing further orders. Please Note: Timeouts will occur after 5 seconds of attempting "
        "to process the bet but please allow up to 15 seconds for a timed out order to appear. After this "
        "time any unprocessed bets will automatically be Lapsed and no longer be available on the Exchange."
    )


class ExecutionReportErrorCode(DocumentedEnum):
    ERROR_IN_MATCHER = doc(
        "The matcher is not healthy. Please note: The error will also be returned if you attempt concurrent "
        "'cancel all' bets requests using cancelOrders which isn't permitted."
    )
    PROCESSED_WITH_ERRORS = doc(
        "The order itself has been accepted, but at least one (possibly all) actions have generated errors."
    )
    BET_ACTION_ERROR = doc(
        "There is an error with an action that has caused the entire order to be rejected. Check the "
        "instructionReports errorCode for the reason for the rejection of the order."
    )
    INVALID_ACCOUNT_STATE = doc("Order rejected due to the account's status (suspended, inactive, dup cards)")
    INVALID_WALLET_STATUS = doc("Order rejected due to the account's wallet's status")
    INSUFFICIENT_FUNDS = doc("Account has exceeded its exposure limit or available to bet limit")
    LOSS_LIMIT_EXCEEDED = doc("The account has exceed the self imposed loss limit")
    MARKET_SUSPENDED = doc("Market is suspended")
    MARKET_NOT_OPEN_FOR_BETTING = doc(
        "Market is not open for betting. It is either not yet active, suspended or closed awaiting settlement."
    )
    DUPLICATE_TRANSACTION = doc(
        "Duplicate customer reference data submitted - Please note: There is a time window associated with "
        "the de-duplication of duplicate submissions which is 60 second"
    )
    INVALID_ORDER = doc(
        "Order cannot be accepted by the matcher due to the combination of actions. For example, bets being "
        "edited are not on the same market, or order includes both edits and placement"
    )
    INVALID_MARKET_ID = doc("Market doesn't exist")
    PERMISSION_DENIED = doc(
        "Business rules do not allow order to be placed. You are either attempting to place the order using "
        "a Delayed Application Key or from a restricted jurisdiction (i.e. USA)"
    )
    DUPLICATE_BETIDS = doc(
        "Duplicate bet ids found. For example, you've included the same betId more than once in a single "
        "cancelOrders request."
    )
    NO_ACTION_REQUIRED = doc("Order hasn't been passed to matcher as system detected there will be no state change")
    SERVICE_UNAVAILABLE = doc("The requested service is unavailable")
    REJECTED_BY_REGULATOR = doc(
        "The regulator rejected the order. On the Italian Exchange this error will occur if more than 50 "
        "bets are sent in a single placeOrders request."
    )
    NO_CHASING = doc(
        "A specific error code that relates to Spanish Exchange markets only which indicates that the bet "
        "placed contravenes the Spanish regulatory rules relating to loss chasing."
    )
    REGULATOR_IS_NOT_AVAILABLE = doc("The underlying regulator service is not available.")
    TOO_MANY_INSTRUCTIONS = doc("The amount of orders exceeded the maximum amount allowed to be executed")
    INVALID_MARKET_VERSION = doc("The supplied market version is invalid. Max length allowed for market version is 12.")
    INVALID_PROFIT_RATIO = doc("The order falls outside the permitted price and size combination.")
    EVENT_EXPOSURE_LIMIT_EXCEEDED = doc("@Deprecated Defined in XML specification but not in API doc")
    EVENT_MATCHED_EXPOSURE_LIMIT_EXCEEDED = doc("@Deprecated Defined in XML specification but not in API doc")
    EVENT_BLOCKED = doc("@Deprecated Defined in XML specification but not in API doc")


class PersistenceType(DocumentedEnum):
    LAPSE = doc("Lapse (cancel) the order automatically when the market is turned in play if the bet is unmatched")
    PERSIST = doc(
        "Persist the unmatched order to in-play. The bet will be placed automatically into the in-play "
        "market at the start of the event. Once in play, the bet won't be cancelled by Betfair if a "
        "material event takes place and will be available until matched or cancelled by the user."
    )
    MARKET_ON_CLOSE = doc("Put the order into the auction (SP) at turn-in-play")


class InstructionReportStatus(DocumentedEnum):
    SUCCESS = doc("The instruction was successful.")
    FAILURE = doc("The instruction failed.")
    TIMEOUT = doc(
        "The order timed out & the status of the bet is unknown. If a TIMEOUT error occurs on a "
        "placeOrders/replaceOrders request, you should check listCurrentOrders to verify the status "
        "of your bets before placing further orders. Please Note: Timeouts will occur after 5 seconds "
        "of attempting to process the bet but please allow up to 15 seconds for a timed out order to "
        "appear. After this time any unprocessed bets will automatically be Lapsed and no longer be "
        "available on the Exchange."
    )


class InstructionReportErrorCode(DocumentedEnum):
    INVALID_BET_SIZE = doc("bet size is invalid for your currency or your regulator")
    INVALID_RUNNER = doc("Runner does not exist, includes vacant traps in greyhound racing")
    BET_TAKEN_OR_LAPSED = doc(
        "Bet cannot be cancelled or modified as it has already been taken or has been cancelled/lapsed "
        "Includes attempts to cancel/modify market on close BSP bets and cancelling limit on close BSP bets. "
        "The error may be returned on placeOrders request if for example a bet is placed at the point when "
        "a market admin event takes place (i.e. market is turned in-play). The error will also be returned "
        "if a market version is submitted and a material change has taken place since the bet was submitted "
        "causing the bet to be rejected."
    )
    BET_IN_PROGRESS = doc("No result was received from the matcher in a timeout configured for the system")
    RUNNER_REMOVED = doc("Runner has been removed from the event")
    MARKET_NOT_OPEN_FOR_BETTING = doc("Attempt to edit a bet on a market that has closed.")
    LOSS_LIMIT_EXCEEDED = doc("The action has caused the account to exceed the self imposed loss limit")
    MARKET_NOT_OPEN_FOR_BSP_BETTING = doc("Market now closed to bsp betting. Turned in-play or has been reconciled")
    INVALID_PRICE_EDIT = doc(
        "Attempt to edit down the price of a bsp limit on close lay bet, or edit up the price of a limit on "
        "close back bet"
    )
    INVALID_ODDS = doc("Odds not on price ladder - either edit or placement")
    INSUFFICIENT_FUNDS = doc(
        "Insufficient funds available to cover the bet action. Either the exposure limit or available to bet "
        "limit would be exceeded."
    )
    INVALID_PERSISTENCE_TYPE = doc("Invalid persistence type for this market, e.g. KEEP for a non in-play market.")
    ERROR_IN_MATCHER = doc("A problem with the matcher prevented this action completing successfully")
    INVALID_BACK_LAY_COMBINATION = doc(
        "The order contains a back and a lay for the same runner at overlapping prices. This would guarantee "
        "a self match. This also applies to BSP limit on close bets."
    )
    ERROR_IN_ORDER = doc("The action failed because the parent order failed")
    INVALID_BID_TYPE = doc("Bid type is mandatory")
    INVALID_BET_ID = doc("Bet for id supplied has not been found")
    CANCELLED_NOT_PLACED = doc("Bet cancelled but replacement bet was not placed")
    RELATED_ACTION_FAILED = doc("Action failed due to the failure of a action on which this action is dependent")
    NO_ACTION_REQUIRED = doc(
        "the action does not result in any state change. eg changing a persistence to it's current value"
    )
    TIME_IN_FORCE_CONFLICT = doc(
        "You may only specify a time in force on either the place request OR on individual limit order "
        "instructions (not both), since the implied behaviors are incompatible."
    )
    UNEXPECTED_PERSISTENCE_TYPE = doc(
        "You have specified a persistence type for a FILL_OR_KILL order, which is nonsensical because no "
        "umatched portion can remain after the order has been placed."
    )
    INVALID_ORDER_TYPE = doc(
        "You have specified a time in force of FILL_OR_KILL, but have included a non-LIMIT order type."
    )
    UNEXPECTED_MIN_FILL_SIZE = doc(
        "You have specified a minFillSize on a limit order, where the limit order's time in force is not "
        "FILL_OR_KILL. Using minFillSize is not supported where the time in force of the request (as "
        "opposed to an order) is FILL_OR_KILL."
    )
    INVALID_CUSTOMER_ORDER_REF = doc("The supplied customer order reference is too long.")
    INVALID_MIN_FILL_SIZE = doc(
        "The minFillSize must be greater than zero and less than or equal to the order's size. The "
        "minFillSize cannot be less than the minimum bet size for your currency."
    )
    BET_LAPSED_PRICE_IMPROVEMENT_TOO_LARGE = doc(
        "Your bet is lapsed. There is better odds than requested available in the market, but your "
        "preferences don't allow the system to match your bet against better odds. Change your betting "
        "preferences to accept better odds if you don't want to receive this error."
    )
    INVALID_CUSTOMER_STRATEGY_REF = doc("@Deprecated Defined in XML specification but not in API doc")
    INVALID_PROFIT_RATIO = doc("The order falls outside the permitted price and size combination.")


class RollupModel(DocumentedEnum):
    STAKE = doc("The volumes will be rolled up to the minimum value which is >= rollupLimit.")
    PAYOUT = doc(
        "The volumes will be rolled up to the minimum value where the payout( price * volume ) "
        "is >= rollupLimit. On a LINE market, volumes will be rolled up where payout( 2.0 * volume ) "
        "is >= rollupLimit"
    )
    MANAGED_LIABILITY = doc(
        "The volumes will be rolled up to the minimum value which is >= rollupLimit, until a lay price "
        "threshold. Thereafter, the volumes will be rolled up to the minimum value such that the liability "
        ">= a minimum liability. Not supported as yet."
    )
    NONE = doc(
        "No rollup will be applied. However, the volumes will be filtered by currency specific minimum "
        "stake unless overridden specifically for the channel."
    )


class GroupBy(DocumentedEnum):
    EVENT_TYPE = doc("A roll up of settled P&L, commission paid and number of bet orders, on a specified event type")
    EVENT = doc("A roll up of settled P&L, commission paid and number of bet orders, on a specified event")
    MARKET = doc("A roll up of settled P&L, commission paid and number of bet orders, on a specified market")
    SIDE = doc(
        "An averaged roll up of settled P&L, and number of bets, on the specified side of a specified "
        "selection within a specified market, that are either settled or voided"
    )
    BET = doc("The P&L, side and regulatory information etc, about each individual bet order.")
    RUNNER = doc("@Deprecated Defined in XML specification but not in API doc")
    STRATEGY = doc("@Deprecated Defined in XML specification but not in API doc")


class BetStatus(DocumentedEnum):
    SETTLED = doc("A matched bet that was settled normally")
    VOIDED = doc("A matched bet that was subsequently voided by Betfair, before, during or after settlement")
    LAPSED = doc("Unmatched bet that was cancelled by Betfair (for example at turn in play).")
    CANCELLED = doc("Unmatched bet that was cancelled by an explicit customer action.")


class TimeInForce(DocumentedEnum):
    FILL_OR_KILL = doc(
        "Execute the transaction immediately and completely (filled to size or between minFillSize and size) "
        "or not at all (cancelled). For LINE markets Volume Weighted Average Price (VWAP) functionality is "
        "disabled."
    )


class BetTargetType(DocumentedEnum):
    BACKERS_PROFIT = doc(
        "The payout requested minus the calculated size at which this LimitOrder is to be placed. "
        "BetTargetType bets are invalid for LINE markets"
    )
    PAYOUT = doc("The total payout requested on a LimitOrder")


class PriceLadderType(DocumentedEnum):
    CLASSIC = doc("Price ladder increments traditionally used for Odds Markets.")
    FINEST = doc("Price ladder with the finest available increment, traditionally used for Asian Handicap markets.")
    LINE_RANGE = doc("Price ladder used for LINE markets. Refer to MarketLineRangeInfo for more details.")


class BetDelayModel(DocumentedEnum):
    PASSIVE = doc(
        "For in-play markets where betDelay > 0, orders that are guaranteed not to match "
        "immediately are accepted straight away, bypassing the bet delay wait."
    )
    DYNAMIC = doc(
        "The DYNAMIC model indicates that a market is subject to dynamic in-play bet delays, "
        "meaning the in-play betDelay value will change while the market is live."
    )


class MarketTypeCode(StrEnum):
    """This is not a complete list, but only a selection of some popular options for direct access.

    A full list can be retrieved from the API using ListMarketTypes.
    """

    ALT_TOTAL_GOALS = auto()
    ANTEPOST_WIN = auto()
    ASIAN_HANDICAP = auto()
    BOTH_TEAMS_TO_SCORE = auto()
    COMBINED_TOTAL = auto()
    CORRECT_SCORE = auto()
    EACH_WAY = auto()
    EXACTA = auto()
    FORECAST = auto()
    HANDICAP = auto()
    MATCH_BET = auto()
    MATCH_ODDS = auto()
    NUMBER_OF_SETS = auto()
    OTHER_PLACE = auto()
    OUTRIGHT_WINNER = auto()
    OVER_UNDER_05 = auto()
    OVER_UNDER_25 = auto()
    PLACE = auto()
    PLAYER_A_WIN_A_SET = auto()
    PLAYER_B_WIN_A_SET = auto()
    QUARTER_WINNER = auto()
    QUINELLA = auto()
    RACE_WIN_DIST = auto()
    SEASON_SPECIALS = auto()
    SET_BETTING = auto()
    SET_CORRECT_SCORE = auto()
    SET_WINNER = auto()
    TOURNAMENT_WINNER = auto()
    TRAP_CHALLENGE = auto()
    UNUSED = auto()
    WIN = auto()
    WITHOUT_FAV = auto()



import msgspec
from msgspec.json import Decoder

from betfair_parser.exceptions import StreamError
from betfair_parser.spec.betting.enums import MarketBettingType, MarketStatus, MarketTypeCode, RunnerStatus  # noqa
from betfair_parser.spec.common import EventTypeIdCode  # noqa for reimport
from betfair_parser.spec.streaming import enums, type_definitions
from betfair_parser.spec.streaming.enums import (
    ChangeType,
    LapseStatusReasonCode,
    MarketDataFilterFields,
    SegmentType,
    StatusErrorCode,
)
from betfair_parser.spec.streaming.messages import (
    MCM,
    OCM,
    Authentication,
    Connection,
    Heartbeat,
    MarketSubscription,
    OrderSubscription,
    Status,
)
from betfair_parser.spec.streaming.type_definitions import (
    LPV,
    PV,
    AvailableToBack,
    AvailableToLay,
    BestAvailableToBack,
    BestAvailableToLay,
    BestDisplayAvailableToBack,
    BestDisplayAvailableToLay,
    KeyLineDefinition,
    KeyLineSelection,
    MarketChange,
    MarketDataFilter,
    MarketDefinition,
    MarketFilter,
    MatchedOrder,
    Order,
    OrderFilter,
    OrderMarketChange,
    OrderRunnerChange,
    RunnerChange,
    RunnerDefinition,
    StartingPriceBack,
    StartingPriceLay,
    StrategyMatchChange,
    StreamRef,
    Trade,
)


StreamRequestType = Authentication | MarketSubscription | OrderSubscription | Heartbeat
StreamResponseType = Connection | Status | MCM | OCM
SubscriptionType = MarketSubscription | OrderSubscription
ChangeMessageType = MCM | OCM
StreamMessageType = StreamResponseType | list[StreamResponseType] | StreamRequestType

_STREAM_DECODER = Decoder(StreamMessageType, strict=False)


def stream_decode(raw: str | bytes) -> StreamMessageType:
    try:
        return _STREAM_DECODER.decode(raw)
    except msgspec.DecodeError as e:
        raise StreamError(str(e))


def stream_decode_lines(raw: str | bytes) -> list[StreamMessageType]:
    try:
        return _STREAM_DECODER.decode_lines(raw)
    except msgspec.DecodeError as e:
        raise StreamError(str(e))



from datetime import datetime
from functools import partial

import msgspec

from betfair_parser.endpoints import SILKS
from betfair_parser.spec.betting.enums import (
    BetDelayModel,
    BetTargetType,
    ExecutionReportErrorCode,
    ExecutionReportStatus,
    InstructionReportErrorCode,
    InstructionReportStatus,
    MarketBettingType,
    MarketStatus,
    MarketTypeCode,
    PersistenceType,
    PriceData,
    PriceLadderType,
    RollupModel,
    RunnerStatus,
    Side,
    TimeInForce,
)
from betfair_parser.spec.common import (
    BaseMessage,
    BetId,
    CompetitionId,
    CountryCode,
    CustomerOrderRef,
    CustomerRef,
    CustomerStrategyRef,
    Date,
    EventId,
    EventTypeId,
    ExchangeId,
    Handicap,
    MarketId,
    MarketType,
    MatchId,
    OrderStatus,
    OrderType,
    Price,
    SelectionId,
    Set,
    Size,
    TimeRange,
    Venue,
    method_tag,
)


betting_tag = partial(method_tag, "SportsAPING/v1.0/")


class Competition(BaseMessage, frozen=True):
    id: CompetitionId | None = None
    name: str | None = None


class CompetitionResult(BaseMessage, frozen=True):
    competition: Competition | None = None
    market_count: int | None = None  # Count of markets associated with this competition
    competition_region: str | None = None  # Region in which this competition is happening


class Event(BaseMessage, frozen=True):
    id: EventId | None = None  # The unique id for the event
    name: str | None = None  # The name of the event
    country_code: CountryCode | None = None  # The ISO-2 code for the event, defaults to GB
    timezone: str | None = None  # The timezone in which the event is taking place
    venue: str | None = None
    open_date: Date | None = None  # The scheduled start date and time of the event


class EventResult(BaseMessage, frozen=True):
    event: Event | None = None
    market_count: int | None = None  # Count of markets associated with this event


class EventType(BaseMessage, frozen=True):
    id: EventTypeId | None = None
    name: str | None = None


class EventTypeResult(BaseMessage, frozen=True):
    event_type: EventType | None = None  # The ID identifying the Event Type
    market_count: int | None = None  # Count of markets associated with this eventType


class MarketTypeResult(BaseMessage, frozen=True):
    market_type: str | None = None
    market_count: int | None = None  # Count of markets associated with this marketType


class CountryCodeResult(BaseMessage, frozen=True):
    country_code: CountryCode | None = None  # The ISO-2 code for the event
    market_count: int | None = None  # Count of markets associated with this Country Code


class VenueResult(BaseMessage, frozen=True):
    venue: Venue | None = None
    market_count: int | None = None  # Count of markets associated with this Venue


class PriceSize(BaseMessage, frozen=True):
    price: Price  # Available price
    size: Size  # Available stake


class TimeRangeResult(BaseMessage, frozen=True):
    time_range: TimeRange | None = None
    market_count: int | None = None  # Count of markets associated with this TimeRange


class MarketFilter(BaseMessage, frozen=True):
    bsp_only: bool | None = None  # Restrict to bsp markets only if True or non-bsp markets if False
    competition_ids: Set[CompetitionId] | None = None  # Restrict markets by the competitions
    event_ids: Set[EventId] | None = None  # Restrict markets by the event id associated with the market
    event_type_ids: Set[EventTypeId] | None = None  # Restrict markets by event type associated with the market
    exchange_ids: Set[ExchangeId] | None = None  # DEPRECATED

    # Restrict to markets that are currently in play if True or are not currently in play if False
    in_play_only: bool | None = None
    market_betting_types: Set[MarketBettingType] | None = None  # Match the betting type of the market
    market_countries: Set[CountryCode] | None = None  # Match the specified country or countries
    market_ids: Set[MarketId] | None = None  # Restrict markets by the market id associated with the market
    market_start_time: TimeRange | None = None  # Restrict to markets with a market start time range
    market_type_codes: Set[MarketTypeCode] | None = None  # Restrict to markets that match the type of the market
    race_types: Set[str] | None = None  # Restrict by race type
    text_query: str | None = None  # Restrict markets by any text associated with the Event name

    # Restrict to markets that will turn in play if True or will not turn in play if False
    turn_in_play_enabled: bool | None = None
    venues: Set[Venue] | None = None  # Restrict markets by the venue associated with the market
    with_orders: Set[str] | None = None  # Markets that have one or more orders of defined OrderStatus


class MarketLineRangeInfo(BaseMessage, frozen=True):
    """Market Line and Range Info"""

    max_unit_value: float  # Maximum value for the outcome, in market units
    min_unit_value: float  # Minimum value for the outcome, in market units
    interval: float  # The odds ladder increment interval
    market_unit: str  # The type of unit the lines are incremented in


class PriceLadderDescription(BaseMessage, frozen=True):
    """Description of the price ladder type and any related data"""

    type: PriceLadderType


class StartingPrices(BaseMessage, frozen=True):
    """Information about the Betfair Starting Price. Only available in BSP markets"""

    near_price: float | None = None  # What the starting price would be if the market was reconciled now
    far_price: float | None = None  # What the starting price would be if the market was reconciled now

    # The total amount of back bets matched at the actual Betfair Starting Price
    back_stake_taken: list[PriceSize] | None = None
    lay_liability_taken: list[PriceSize] | None = None  # The lay amount matched at the actual Betfair Starting Price
    actual_sp: float | None = msgspec.field(name="actualSP", default=None)  # The final BSP price for this runner


class ExchangePrices(BaseMessage, frozen=True):
    available_to_back: list[PriceSize] | None = None
    available_to_lay: list[PriceSize] | None = None
    traded_volume: list[PriceSize] | None = None


class Order(BaseMessage, frozen=True):
    bet_id: BetId
    order_type: OrderType
    status: OrderStatus
    persistence_type: PersistenceType
    side: Side  # Side (BACK or LAY)
    price: Price
    size: Size
    bsp_liability: Size  # Liability of a given BSP bet
    placed_date: Date  # Date and time the bet was placed
    avg_price_matched: Price | None = None  # Average price matched at
    size_matched: Size | None = None  # Current amount of this bet that was matched
    size_remaining: Size | None = None  # Current amount of this bet that is unmatched
    size_lapsed: Size | None = None  # Current amount of this bet that was lapsed
    size_cancelled: Size | None = None  # Current amount of this bet that was cancelled
    size_voided: Size | None = None  # Current amount of this bet that was voided
    customer_order_ref: CustomerOrderRef | None = None  # Customer Order Reference
    customer_strategy_ref: CustomerStrategyRef | None = None  # Customer Strategy Reference


class Match(BaseMessage, kw_only=True, frozen=True):
    """An individual bet Match, or rollup by price or avg price.

    Rollup depends on the requested MatchProjection.
    """

    bet_id: BetId | None = None  # Bet ID (present if no rollup)
    match_id: MatchId | None = None  # Match ID (present if no rollup)
    side: Side  # Side (BACK or LAY)
    price: Price  # Match price
    size: Size  # Size matched at
    match_date: Date | None = None  # Match date (present if no rollup)


class MarketVersion(BaseMessage, frozen=True):
    version: int | None = None  # A non-monotonically increasing number indicating market changes


class MarketRates(BaseMessage, frozen=True):
    market_base_rate: float  # Market base rate
    discount_allowed: bool  # Indicates whether discount is allowed on this market


class MarketLicence(BaseMessage, frozen=True):
    wallet: str  # Wallet from which funds will be taken when betting on this market
    rules: str | None = None
    rules_has_date: bool | None = None  # Markets start date and time are relevant to the rules
    clarifications: str | None = None  # Clarifications to the rules for the market


class MarketDescription(BaseMessage, kw_only=True, frozen=True):
    betting_type: MarketBettingType
    bsp_market: bool  # Indicates if the market supports Betfair SP betting
    clarifications: str | None = None  # Additional information regarding the market
    discount_allowed: bool  # Indicates whether user's discount rate is taken into account on this
    each_way_divisor: float | None = None  # Divisor for EACH_WAY market type
    line_range_info: MarketLineRangeInfo | None = None  # Line range info for line markets
    market_base_rate: float  # Commission rate applicable to the market
    market_time: Date  # Scheduled start time of the market
    market_type: str  # Market base type
    persistence_enabled: bool  # Indicates if the market supports 'Keep' bets if turned in-play
    price_ladder_description: PriceLadderDescription | None = None  # Details about the price ladder in use
    race_type: str | None = None  # External identifier of a race type
    regulator: str  # Market regulator
    rules: str | None = None  # The market rules
    rules_has_date: bool | None = None  # Indicates whether rules have a date included
    settle_time: Date | None = None
    suspend_time: Date  # Next time the market will be suspended for betting, usually just marketTime
    turn_in_play_enabled: bool  # Indicates if the market is set to turn in-play
    wallet: str | None = None  # The wallet to which the market belongs


# TODO: Some fields in the meta data should be country codes. Unfortunately, sometimes they contain
#       erroneous data that fails verification. This should be switched to CountryCode as soon as there is
#       some more fine-grained error handling possible in msgspec. Related issue:
#       https://github.com/jcrist/msgspec/issues/420
_MetaCountryCode = str  # CountryCode


class RunnerMetaData(BaseMessage, frozen=True, rename="upper"):
    # Yes, this is the only type definition, that has (mostly) upper-case key names
    """
    Runner metadata as defined in the API as additional information.
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Additional+Information
    """

    adjusted_rating: int | None = None  # Race-specific ratings that reflect weights allocated in the race
    age: int | None = None  # The age of the horse
    bred: _MetaCountryCode | None = None  # The country in which the horse was born
    cloth_number: int | None = None  # The number on the saddle-cloth
    cloth_number_alpha: str | None = None  # The number on the saddle cloth for US paired runners, e.g. "1A"
    colour_type: str | None = None  # The colour of the horse
    colours_description: str | None = None  # The textual description of the jockey silk
    colours_filename: str | None = None  # Image representing the jockey silk
    dam_bred: _MetaCountryCode | None = None  # The country where the horse's mother was born
    dam_name: str | None = None  # The name of the horse's mother
    dam_year_born: int | None = None  # The year the horse’s mother's birth
    damsire_bred: _MetaCountryCode | None = None  # The country where the horse's grandfather was born
    damsire_name: str | None = None  # The name of the horse's grandfather
    damsire_year_born: int | None = None  # Year in which the horse's grandfather was born on its mother's side
    days_since_last_run: int | None = None  # The number of days since the horse last ran
    forecastprice_denominator: int | None = None  # The forecast price denominator
    forecastprice_numerator: int | None = None  # The forecast price numerator
    form: str | None = None  # The horses recent form
    jockey_claim: int | None = None  # Reduction in the weight that the horse carries for a particular jockey
    jockey_name: str | None = None  # Name of the jockey. This field will contain 'Reserve' if it's a reserve runner
    official_rating: int | None = None  # The horses official rating
    owner_name: str | None = None  # The owner of the horse
    runner_id: int | None = msgspec.field(name="runnerId", default=None)  # The runnerId for the horse
    sex_type: str | None = None  # The sex of the horse
    sire_bred: _MetaCountryCode | None = None  # The country where the horse's father was bred
    sire_name: str | None = None  # The name of the horse's father
    sire_year_born: int | None = None  # The year the horse's father was born
    stall_draw: int | None = None  # The stall number the horse is starting from
    trainer_name: str | None = None  # The name of the horse's trainer
    wearing: str | None = None  # Any extra equipment the horse is wearing
    weight_units: str | None = None  # The unit of weight used.
    weight_value: float | None = None  # The weight of the horse

    def __post_init__(self):
        force_setattr = msgspec.structs.force_setattr
        cur_year = datetime.now().year
        if self.weight_value is not None and self.weight_value <= 0:
            force_setattr(self, "weight_value", None)
        if self.stall_draw is not None and not 0 < self.stall_draw < 50:
            force_setattr(self, "stall_draw", None)
        if self.sire_year_born is not None and not cur_year - 40 < self.sire_year_born < cur_year:
            force_setattr(self, "sire_year_born", None)
        if self.dam_year_born is not None and not cur_year - 40 < self.dam_year_born < cur_year:
            force_setattr(self, "dam_year_born", None)
        if self.damsire_year_born is not None and not cur_year - 60 < self.damsire_year_born < cur_year:
            force_setattr(self, "damsire_year_born", None)
        if self.cloth_number is not None and not 0 < self.cloth_number < 50:
            force_setattr(self, "cloth_number", None)
        if self.age is not None and not 1 < self.age < 30:
            force_setattr(self, "age", None)

    @property
    def colours_url(self):
        if not self.colours_filename:
            return None
        return SILKS + self.colours_filename

    @property
    def forecastprice_decimal(self):
        if not self.forecastprice_numerator or not self.forecastprice_denominator:
            return None
        return self.forecastprice_numerator / self.forecastprice_denominator + 1


class RunnerCatalog(BaseMessage, frozen=True):
    """Information about the Runners (selections) in a market"""

    selection_id: SelectionId  # The unique id for the selection
    runner_name: str  # The name of the runner

    # The handicap applies to market with the MarketBettingType ASIAN_HANDICAP_SINGLE_LINE
    # & ASIAN_HANDICAP_DOUBLE_LINE only
    handicap: Handicap
    sort_priority: int | None = None  # This is marked as REQUIRED in the API doc, but omitted sometimes
    metadata: RunnerMetaData | None = None  # Metadata associated with the runner

    @property
    def name(self):
        return self.runner_name


class Runner(BaseMessage, frozen=True):
    """The dynamic data about runners in a market"""

    selection_id: SelectionId  # The unique id of the runner (selection)
    handicap: Handicap
    status: RunnerStatus  # The status of the selection
    adjustment_factor: float | None = None  # The adjustment factor applied if the selection is removed
    last_price_traded: float | None = None  # The price of the most recent bet matched on this selection
    total_matched: float | None = None  # The total amount matched on this runner
    removal_date: Date | None = None  # If date and time the runner was removed
    sp: StartingPrices | None = None  # The BSP related prices for this runner
    ex: ExchangePrices | None = None  # The Exchange prices available for this runner
    orders: list[Order] | None = None  # List of orders in the market
    matches: list[Match] | None = None  # List of matches (i.e., orders that have been fully or partially executed)
    matches_by_strategy: dict[str, list[Match]] | None = None  # All matches for each strategy, sort by matched data


class MarketCatalogue(BaseMessage, frozen=True):
    market_id: MarketId  # The unique identifier for the market
    market_name: str  # The name of the market
    market_start_time: Date | None = None  # Only returned when the MARKET_START_TIME enum is requested
    total_matched: float | None = None  # The total amount of money matched on the market
    event_type: EventType | None = None  # The Event Type the market is contained within
    competition: Competition | None = None  # The competition the market is contained within
    description: MarketDescription | None = None  # Details about the market
    event: Event | None = None  # The event the market is contained within
    runners: list[RunnerCatalog] | None = None  # The runners (selections) contained in the market


class KeyLineSelection(BaseMessage, frozen=True):
    """Description of a market's key line selection"""

    selection_id: SelectionId  # Selection ID of the runner in the key line handicap
    handicap: Handicap  # Handicap value of the key line


class KeyLineDescription(BaseMessage, frozen=True):
    """A list of KeyLineSelection objects describing the key line for the market"""

    key_line: list[KeyLineSelection]  # A list of KeyLineSelection objects


class MarketBook(BaseMessage, frozen=True):
    """The dynamic data in a market"""

    market_id: MarketId  # The unique identifier for the market
    is_market_data_delayed: bool  # True if the data returned by listMarketBook will be delayed
    bet_delay: int | None = None  # The number of seconds an order is held until it is submitted into the market
    bsp_reconciled: bool | None = None  # True if the market starting price has been reconciled
    complete: bool | None = None  # If false, runners may be added to the market
    cross_matching: bool | None = None  # True if cross-matching is enabled for this market
    inplay: bool | None = None  # True if the market is currently in play
    key_line_description: KeyLineDescription | None = None  # Description of a market's key line
    last_match_time: Date | None = None  # The most recent time an order was executed
    number_of_active_runners: int | None = None  # The number of runners that are currently active
    number_of_runners: int | None = None  # The number of runners in the market
    number_of_winners: int | None = None  # The number of selections that could be settled as winners
    runners_voidable: bool | None = None  # True if runners in the market can be voided
    status: MarketStatus | None = None  # The status of the market
    total_available: float | None = None  # The total amount of orders that remain unmatched
    total_matched: float | None = None  # The total amount matched on the market
    version: int | None = None  # The version of the market
    runners: list[Runner] | None = None  # Information about the runners (selections) in the market
    bet_delay_models: list[BetDelayModel] | None = None  # Indicates which bet delay models are applied to a market


class ItemDescription(BaseMessage, frozen=True):
    """
    This object contains some text which may be useful to render a betting history view.
    It offers no long-term warranty as to the correctness of the text.
    """

    event_type_desc: str | None = None  # The event type name translated into the requested locale
    event_desc: str | None = None  # The event name or openDate + venue translated into the requested locale
    market_desc: str | None = None  # The market name or racing market type translated into the requested locale
    market_type: MarketType | None = None  # The market type e.g. MATCH_ODDS, PLACE, WIN etc.
    market_start_time: Date | None = None  # The start time of the market in ISO-8601 format, not translated
    runner_desc: str | None = None  # The runner name translated into the requested locale
    number_of_winners: int | None = None  # The number of winners on a market
    each_way_divisor: float | None = None  # The divisor for the EACH_WAY market type


class ClearedOrderSummary(BaseMessage, frozen=True):
    """Summary of a cleared order"""

    event_type_id: EventTypeId | None = None  # The id of the event type bet on
    event_id: EventId | None = None  # The id of the event bet on
    market_id: MarketId | None = None  # The id of the market bet on
    selection_id: SelectionId | None = None  # The id of the selection bet on
    handicap: Handicap | None = None  # The handicap value for Asian handicap markets
    bet_id: BetId | None = None  # The id of the bet
    placed_date: Date | None = None  # The date the bet order was placed by the customer
    persistence_type: PersistenceType | None = None  # The turn in play persistence state of the order
    order_type: OrderType | None = None
    side: Side | None = None  # Whether the bet was a back or lay bet
    item_description: ItemDescription | None = None  # Container for ancillary data and localized text
    bet_outcome: str | None = None  # The settlement outcome of the bet
    price_requested: Price | None = None  # The average requested price across all settled bet orders under this item
    settled_date: Date | None = None  # The date and time the bet order was settled by Betfair
    last_matched_date: Date | None = None  # The date and time the last bet order was matched by Betfair
    bet_count: int | None = None  # The number of actual bets within this grouping
    commission: Size | None = None  # Cumulative commission paid by the customer across all bets under this item
    price_matched: Price | None = None  # The average matched price across all settled bets or bet fragments
    price_reduced: bool | None = None  # Indicates if the matched price was affected by a reduction factor
    size_settled: Size | None = None  # The cumulative bet size that was settled as matched or voided under this item
    profit: float | None = None  # The profit or loss gained on this line
    size_cancelled: Size | None = None  # The amount of the bet that was cancelled
    customer_order_ref: CustomerOrderRef | None = None  # Defined by the customer for the bet order
    customer_strategy_ref: CustomerStrategyRef | None = None  # Defined by the customer for the bet order


class ClearedOrderSummaryReport(BaseMessage, frozen=True):
    """A container representing search results"""

    cleared_orders: list[ClearedOrderSummary]  # The list of cleared orders returned by the query
    more_available: bool  # Indicates whether there are further result items beyond this page


class RunnerId(BaseMessage, frozen=True):
    """Unique identifier for a runner"""

    market_id: MarketId  # The id of the market bet on
    selection_id: SelectionId  # The id of the selection bet on
    handicap: Handicap | None = None  # The handicap associated with the runner in case of Asian handicap markets


class CurrentItemDescription(BaseMessage, frozen=True):
    """This object contains ancillary information about the item"""

    market_version: MarketVersion  # The relevant version of the market for this item


class CurrentOrderSummary(BaseMessage, frozen=True):
    """Summary of a current order"""

    bet_id: BetId  # The bet ID of the original place order
    market_id: MarketId  # The market ID the order is for
    selection_id: SelectionId  # The selection ID the order is for
    handicap: Handicap  # The handicap associated with the runner in case of Asian handicap markets
    price_size: PriceSize  # The price and size of the bet
    bsp_liability: Size  # The liability of a given BSP bet
    side: Side  # BACK/LAY
    status: OrderStatus  # The status of the order
    persistence_type: PersistenceType  # What to do with the order at turn-in-play
    order_type: OrderType  # BSP Order type
    placed_date: Date  # The date the bet was placed

    # Date of the last matched bet fragment.
    # Mandatory according to documentation, but optional in reality
    matched_date: Date | None = None
    average_price_matched: Price | None = None  # The average price matched at
    size_matched: Size | None = None
    size_remaining: Size | None = None
    size_lapsed: Size | None = None
    size_cancelled: Size | None = None
    size_voided: Size | None = None
    regulator_auth_code: str | None = None
    regulator_code: str | None = None
    customer_order_ref: str | None = None  # The order reference defined by the customer for this bet
    customer_strategy_ref: str | None = None  # The strategy reference defined by the customer for this bet
    current_item_description: CurrentItemDescription | None = None  # Container for ancillary data for this item


class CurrentOrderSummaryReport(BaseMessage, frozen=True):
    """A container representing search results"""

    current_orders: list[CurrentOrderSummary]  # The list of current orders returned by the query
    more_available: bool  # Indicates whether there are further result items beyond this page


class LimitOrder(BaseMessage, frozen=True):
    """Place a new LIMIT order (simple exchange bet for immediate execution)"""

    size: Size
    price: Price
    persistence_type: PersistenceType | None = None
    time_in_force: TimeInForce | None = None
    min_fill_size: Size | None = None
    bet_target_type: BetTargetType | None = None
    bet_target_size: Size | None = None


class LimitOnCloseOrder(BaseMessage, frozen=True):
    liability: Size
    price: Price


class MarketOnCloseOrder(BaseMessage, frozen=True):
    liability: Size


class PlaceInstruction(BaseMessage, kw_only=True, frozen=True):
    """Instruction to place a new order"""

    order_type: OrderType  # The order type
    selection_id: SelectionId  # The selection ID
    handicap: Handicap | None = None  # The handicap associated with the runner in case of Asian handicap markets
    side: Side  # Back or Lay
    limit_order: LimitOrder | None = None  # A simple exchange bet for immediate execution

    # Bets matched if the returned starting price is better than a specified price
    limit_on_close_order: LimitOnCloseOrder | None = None

    # Bets matched and settled at a price representative of the market at the point it turns in-play
    market_on_close_order: MarketOnCloseOrder | None = None
    customer_order_ref: str | None = None  # An optional reference to identify instructions


class PlaceInstructionReport(BaseMessage, kw_only=True, frozen=True):
    """Report for a place instruction"""

    status: InstructionReportStatus  # The instruction report status
    error_code: InstructionReportErrorCode | None = None
    order_status: OrderStatus | None = None
    instruction: PlaceInstruction  # The place instruction
    bet_id: BetId | None = None  # The bet ID of the placed order, if successful
    placed_date: Date | None = None  # The date and time the bet was placed, if successful
    average_price_matched: Price | None = None  # The average price matched at, if successful
    size_matched: Size | None = None  # The current amount of the bet that was matched, if successful


class PlaceExecutionReport(BaseMessage, kw_only=True, frozen=True):
    customer_ref: CustomerRef | None = None  # Echo of the customer reference if passed
    status: ExecutionReportStatus  # The execution report status
    error_code: ExecutionReportErrorCode | None = None  # The execution report error code
    market_id: MarketId | None = None  # Echo of the market ID passed
    instruction_reports: list[PlaceInstructionReport] | None = None  # The list of place instruction reports


class CancelInstruction(BaseMessage, frozen=True):
    """Instruction to fully or partially cancel an order (only applies to LIMIT orders)"""

    bet_id: BetId
    # If supplied then this is a partial cancel. Should be set to 'null' if no size
    # reduction is required.
    size_reduction: Size | None = None


class ReplaceInstruction(BaseMessage, frozen=True):
    """Instruction to replace a LIMIT or LIMIT_ON_CLOSE order at a new price."""

    bet_id: BetId  # Unique identifier for the bet
    new_price: Price  # The price to replace the bet at


class CancelInstructionReport(BaseMessage, kw_only=True, frozen=True):
    status: InstructionReportStatus  # Whether the command succeeded or failed
    error_code: InstructionReportErrorCode | None = None  # Cause of failure, or null if command succeeds
    instruction: CancelInstruction | None = None  # The instruction that was requested
    size_cancelled: float | None = None  # The API states, that this is mandatory, but it's skipped in case of error
    cancelled_date: Date | None = None  # The API states, that this is mandatory, but it's skipped in case of error


class CancelExecutionReport(BaseMessage, kw_only=True, frozen=True):
    customer_ref: CustomerRef | None = None  # Echo of the customerRef if passed
    status: ExecutionReportStatus
    error_code: ExecutionReportErrorCode | None = None
    market_id: MarketId | None = None  # Echo of marketId passed
    instruction_reports: list[CancelInstructionReport] | None = None


class ReplaceInstructionReport(BaseMessage, frozen=True):
    status: InstructionReportStatus  # Whether the command succeeded or failed
    error_code: InstructionReportErrorCode | None = None  # Cause of failure, or null if command succeeds
    cancel_instruction_report: CancelInstructionReport | None = None  # Cancellation report for the original order
    place_instruction_report: PlaceInstructionReport | None = None  # Placement report for the new order


class ReplaceExecutionReport(BaseMessage, kw_only=True, frozen=True):
    customer_ref: CustomerRef | None = None  # Echo of the customerRef if passed.
    status: ExecutionReportStatus
    error_code: ExecutionReportErrorCode | None = None
    market_id: MarketId | None = None  # Echo of marketId passed
    instruction_reports: list[ReplaceInstructionReport] | None = None


class UpdateInstruction(BaseMessage, frozen=True):
    """Instruction to update LIMIT bet's persistence of an order that do not affect exposure"""

    bet_id: BetId  # Unique identifier for the bet
    new_persistence_type: PersistenceType  # The new persistence type to update this bet to


class UpdateInstructionReport(BaseMessage, kw_only=True, frozen=True):
    status: InstructionReportStatus  # Whether the command succeeded or failed
    error_code: InstructionReportErrorCode | None = None  # Cause of failure, or null if command succeeds
    instruction: UpdateInstruction  # The instruction that was requested


class UpdateExecutionReport(BaseMessage, frozen=True):
    customer_ref: CustomerRef | None  # Echo of the customerRef if passed.
    status: ExecutionReportStatus
    error_code: ExecutionReportErrorCode | None
    market_id: MarketId | None  # Echo of marketId passed
    instruction_reports: list[UpdateInstructionReport] | None


class ExBestOffersOverrides(BaseMessage, frozen=True):
    """Options to alter the default representation of best offer prices"""

    best_prices_depth: int | None = None  # The maximum number of prices to return on each side for each runner
    rollup_model: RollupModel | None = None  # The model to use when rolling up available sizes
    rollup_limit: int | None = None  # The volume limit to use when rolling up returned sizes
    rollup_liability_threshold: float | None = None  # Only applicable when rollupModel is MANAGED_LIABILITY
    rollup_liability_factor: int | None = None  # Only applicable when rollupModel is MANAGED_LIABILITY


class PriceProjection(BaseMessage, frozen=True):
    """Selection criteria of the returning price data"""

    price_data: Set[PriceData] | None = None  # The basic price data you want to receive in the response

    # Options to alter the default representation of best offer prices
    ex_best_offers_overrides: ExBestOffersOverrides | None = None
    virtualise: bool | None = None  # Indicates if the returned prices should include virtual prices

    # Indicates if the volume returned at each price point should be the absolute value or a
    # cumulative sum of volumes available at the price and all better prices. If unspecified
    # defaults to false. Applicable to EX_BEST_OFFERS and EX_ALL_OFFERS price projections.
    # Not supported as yet.
    rollover_stakes: bool | None = None


class RunnerProfitAndLoss(BaseMessage, frozen=True):
    """Profit and loss if selection wins or loses"""

    selection_id: SelectionId | None = None  # The unique identifier for the selection
    if_win: float | None = None  # Profit or loss for the market if this selection is the winner
    if_lose: float | None = None  # Profit or loss for the market if this selection is the loser

    # Profit or loss for the market if this selection is placed (applies to marketType EACH_WAY only)
    if_place: float | None = None


class MarketProfitAndLoss(BaseMessage, frozen=True):
    """Profit and loss in a market"""

    market_id: MarketId | None = None  # The unique identifier for the market
    commission_applied: float | None = None  # The commission rate applied to P&L values
    profit_and_losses: list[RunnerProfitAndLoss] | None = None  # Calculated profit and loss data



from betfair_parser.spec.betting import enums, operations, type_definitions
from betfair_parser.spec.betting.enums import (
    BetStatus,
    BetTargetType,
    ExecutionReportErrorCode,
    ExecutionReportStatus,
    GroupBy,
    InstructionReportErrorCode,
    InstructionReportStatus,
    MarketBettingType,
    MarketProjection,
    MarketSort,
    MarketStatus,
    MarketTypeCode,
    MatchProjection,
    OrderBy,
    OrderProjection,
    PersistenceType,
    PriceData,
    PriceLadderType,
    RollupModel,
    RunnerStatus,
    Side,
    SortDir,
    TimeGranularity,
    TimeInForce,
)
from betfair_parser.spec.betting.listings import (
    ListCompetitions,
    ListCountries,
    ListEvents,
    ListEventTypes,
    ListMarketBook,
    ListMarketCatalog,
    ListMarketCatalogue,
    ListMarketProfitAndLoss,
    ListMarketTypes,
    ListRunnerBook,
    ListTimeRanges,
    ListVenues,
)
from betfair_parser.spec.betting.orders import (
    CancelOrders,
    ListClearedOrders,
    ListCurrentOrders,
    PlaceOrders,
    ReplaceOrders,
    UpdateOrders,
)
from betfair_parser.spec.betting.type_definitions import (
    CancelExecutionReport,
    CancelInstruction,
    CancelInstructionReport,
    ClearedOrderSummary,
    ClearedOrderSummaryReport,
    Competition,
    CompetitionResult,
    CountryCodeResult,
    CurrentItemDescription,
    CurrentOrderSummary,
    CurrentOrderSummaryReport,
    Event,
    EventResult,
    EventType,
    EventTypeResult,
    ExBestOffersOverrides,
    ExchangePrices,
    ItemDescription,
    KeyLineDescription,
    KeyLineSelection,
    LimitOnCloseOrder,
    LimitOrder,
    MarketBook,
    MarketCatalogue,
    MarketDescription,
    MarketFilter,
    MarketLicence,
    MarketLineRangeInfo,
    MarketOnCloseOrder,
    MarketRates,
    MarketTypeResult,
    MarketVersion,
    Match,
    Order,
    PlaceExecutionReport,
    PlaceInstruction,
    PlaceInstructionReport,
    PriceLadderDescription,
    PriceProjection,
    PriceSize,
    ReplaceExecutionReport,
    ReplaceInstruction,
    ReplaceInstructionReport,
    Runner,
    RunnerCatalog,
    RunnerMetaData,
    StartingPrices,
    TimeRangeResult,
    UpdateInstruction,
    VenueResult,
)
from betfair_parser.spec.common import EventTypeIdCode, OrderStatus, OrderType, TimeRange  # noqa for reimport



from collections import defaultdict
from dataclasses import dataclass


ACCOUNTS = "https://api.betfair.com/exchange/account/json-rpc/v1/"
BETTING = "https://api.betfair.com/exchange/betting/json-rpc/v1/"
SCORES = "https://api.betfair.com/exchange/scores/json-rpc/v1/"
STREAM = "ndjson://stream-api.betfair.com:443"
STREAM_INTEGRATION = "ndjson://stream-api-integration.betfair.com:443"
SILKS = "https://content-cache.cdnppb.net/feeds_images/Horses/SilkColours/"

_IDENTITY = "https://identitysso.betfair{tld}/api/"
_IDENTITY_CERT = "https://identitysso-cert.betfair{tld}/api/"
_NAVIGATION = "https://api.betfair{tld}/exchange/betting/rest/v1/{locale}/navigation/menu.json"
_HEARTBEAT = "https://api.betfair{tld}/exchange/heartbeat/json-rpc/v1/"

IDENTITY = defaultdict(
    lambda: _IDENTITY.format(tld=".com"),
    ESP=_IDENTITY.format(tld=".es"),
    ITA=_IDENTITY.format(tld=".it"),
    ROU=_IDENTITY.format(tld=".ro"),
    SWE=_IDENTITY.format(tld=".se"),
    AUS=_IDENTITY.format(tld=".com.au"),
)
IDENTITY_CERT = defaultdict(
    lambda: _IDENTITY_CERT.format(tld=".com"),
    ESP=_IDENTITY_CERT.format(tld=".es"),
    ITA=_IDENTITY_CERT.format(tld=".it"),
    ROU=_IDENTITY_CERT.format(tld=".ro"),
    SWE=_IDENTITY_CERT.format(tld=".se"),
)
NAVIGATION = defaultdict(
    lambda: _NAVIGATION.format(tld=".com", locale="en"),
    ESP=_NAVIGATION.format(tld=".es", locale="es"),
    ITA=_NAVIGATION.format(tld=".it", locale="it"),
    DEU=_NAVIGATION.format(tld=".com", locale="de"),
    SWE=_NAVIGATION.format(tld=".com", locale="sv"),
    PRT=_NAVIGATION.format(tld=".com", locale="pt"),
    RUS=_NAVIGATION.format(tld=".com", locale="ru"),
    DNK=_NAVIGATION.format(tld=".com", locale="da"),
)
HEARTBEAT = defaultdict(
    lambda: _HEARTBEAT.format(tld=".com", locale="en"),
    ESP=_HEARTBEAT.format(tld=".es", locale="es"),
    ITA=_HEARTBEAT.format(tld=".it", locale="it"),
)


@dataclass(frozen=True)
class EndpointConfig:
    identity: str
    identity_cert: str
    navigation: str
    heartbeat: str
    stream: str
    accounts = ACCOUNTS
    betting = BETTING
    scores = SCORES

    def for_request(self, req):
        return getattr(self, req.endpoint_type.value)

    def url_for_request(self, req):
        return f"{self.for_request(req)}{req.method}"


def endpoint(country_code="GBR", integration=False):
    return EndpointConfig(
        identity=IDENTITY[country_code],
        identity_cert=IDENTITY_CERT[country_code],
        navigation=NAVIGATION[country_code],
        heartbeat=HEARTBEAT[country_code],
        stream=STREAM_INTEGRATION if integration else STREAM,
    )


ENDPOINTS = endpoint()



from betfair_parser.strenums import DocumentedEnum, doc


class SubscriptionStatus(DocumentedEnum):
    ALL = doc("Any subscription status")
    ACTIVATED = doc("Only activated subscriptions")
    UNACTIVATED = doc("Only unactivated subscriptions")
    CANCELLED = doc("Only cancelled subscriptions")
    EXPIRED = doc("Only expired subscriptions")


class Status(DocumentedEnum):
    SUCCESS = doc("Success status")


class ItemClass(DocumentedEnum):
    UNKNOWN = doc(
        "Statement item not mapped to a specific class. All values will be concatenated "
        "into a single key/value pair.The key will be 'unknownStatementItem' and the "
        "value will be a comma separated string.Please note: This is used to represent "
        "commission payment items."
    )


class Wallet(DocumentedEnum):
    UK = doc("The Global Exchange wallet")
    AUSTRALIAN = doc("@Deprecated but still mentioned in the XML specification")


class IncludeItem(DocumentedEnum):
    ALL = doc("Include all items")
    DEPOSITS_WITHDRAWALS = doc("Include payments only")
    EXCHANGE = doc("Include exchange bets only")
    POKER_ROOM = doc("Include poker transactions only")


class WinLose(DocumentedEnum):
    RESULT_ERR = doc(
        "Record has been affected by a unsettlement. There is no impact on the balance "
        "for these records, this is just a label to say that these are to be corrected."
    )
    RESULT_FIX = doc(
        "Record is a correction to the balance to reverse the impact of records shown "
        "as in error. If commission has been paid on the original settlement then there "
        "will be a second FIX record to reverse the commission."
    )
    RESULT_LOST = doc("Loss")
    RESULT_NOT_APPLICABLE = doc("Amounts relating to commission payments.")
    RESULT_WON = doc("Won")
    COMMISSION_REVERSAL = doc(
        "Betfair has restored the funds to your account that it previously received from you in commission."
    )


class GrantType(DocumentedEnum):
    AUTHORIZATION_CODE = doc(
        "Returned via the Vendor Web API token request. The authorization code will be valid "
        "for a single use for 10 minutes."
    )
    REFRESH_TOKEN = doc("A token that can be used to create a new access token when using the Vendor Web API")


class TokenType(DocumentedEnum):
    BEARER = doc("Token type used for Vendor Web API interactions for making requests on a customers behalf.")


class AffiliateRelationStatus(DocumentedEnum):
    INVALID_USER = doc("Provided vendor client ID is not valid")
    AFFILIATED = doc("Vendor client ID valid and affiliated")
    NOT_AFFILIATED = doc("Vendor client ID valid but not affiliated")


class MarketType(DocumentedEnum):
    """Legacy data"""

    A = doc("Asian Handicap")
    L = doc("Line market")
    O = doc("Odds market")  # noqa
    R = doc("Range market.")
    NOT_APPLICABLE = doc("The market does not have an applicable marketType.")
